package jimm.datavision;
import jimm.datavision.source.Column;
import java.util.*;

/**
 * A group uses a data source column or user column to define record
 * grouping. A group may contain multiple header and footer sections.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Group implements Writeable {

/**
 * Sort records by the group's column in ascending order (the default).
 */
public static final int SORT_ASCENDING = 0;
/**
 * Sort records by the group's column in descending order.
 */
public static final int SORT_DESCENDING = 1;

protected Report report;
protected ArrayList headerSections;
protected ArrayList footerSections;
protected Column column;
protected int sortOrder;
protected Object value;		// Current row value
protected boolean newValue;
protected boolean firstValue;
protected int recordCount;

/**
 * Creates a new group and gives it a header section containing a column
 * field and an empty footer section.
 *
 * @param report a report
 * @param col a database column
 * @return a new group with a header section containing a column field and
 * an empty footer section
 */
public static Group create(Report report, Column col) {
    Group group = new Group(report, col);

    // Create new header section to add to group.
    Section header = new Section(report, I18N.get("Report.group_header"),
				 false);

    // Create field for selected column. Make it bold.
    Field f = Field.create(null, report, header, "column", col.getId(), true);
    Format fmt = new Format();
    fmt.setBold(true);
    f.setFormat(fmt);

    // Add field to header section and add section to group.
    header.addField(f);
    group.addHeader(header);

    // Add empty footer section.
    group.addFooter(new Section(report, I18N.get("Report.group_footer"),
				false));

    return group;
}

/**
 * Constructor.
 *
 * @param report a report
 * @param col database column
 */
public Group(Report report, Column col) {
    this.report = report;
    column = col;
    sortOrder = SORT_ASCENDING;
    headerSections = new ArrayList();
    footerSections = new ArrayList();
}

/**
 * Returns the column used by this group.
 *
 * @return a column
 */
public Column getColumn() { return column; }

/**
 * Sets the column used by this group.
 *
 * @param newColumn the new column
 */
public void setColumn(Column newColumn) { column = newColumn; }

/**
 * Returns the sort order (either <code>SORT_ASCENDING</code> or
 * <code>SORT_DESCENDING</code>).
 *
 * @return either <code>SORT_ASCENDING</code> or <code>SORT_DESCENDING</code>
 */
public int getSortOrder() { return sortOrder; }

/**
 * Sets the sort order.
 *
 * @param newSortOrder either <code>SORT_ASCENDING</code> or
 * <code>SORT_DESCENDING</code>
 */
public void setSortOrder(int newSortOrder) { sortOrder = newSortOrder; }

/**
 * Returns the value of this group's column. Only valid while
 * the report is running.
 *
 * @return the value of the column; undefined when the report
 * is not running
 */
public Object getValue() { return value; }

/**
 * Sets the group value that is returned by <code>getValue</code>. This
 * method should only be called by the report while it is running.
 *
 * @param report the report from which we retrieve our column's value
 */
public void setValue(Report report) {
    Object val = report.columnValue(column);
    if (value == null) {
	value = val;
	firstValue = true;
	newValue = true;
    }
    else if (value.equals(val)) {
	newValue = false;
	firstValue = false;
    }
    else {
	value = val;
	newValue = true;
	firstValue = false;
    }
}

public void updateCounter() {
    if (newValue)
	recordCount = 1;
    else
	++recordCount;
}

/**
 * Returns <code>true</code> when the value of the column has
 * changed.
 *
 * @return <code>true</code> when the value of the column has
 * changed 
 */
public boolean isNewValue() { return newValue; }

/**
 * Returns the number of records in the group so far.
 *
 * @return the number of records in the group so far
 */
public int getRecordCount() { return recordCount; }

/**
 * Layout engines need to call this method when a group's footer is being
 * output not because this group has a new value but because some previous
 * group's value changed and we want to output this group's footer.
 *
 * @see LayoutEngine#groupFooters
 */
void forceFooterOutput() { newValue = true; }

/**
 * Returns <code>true</code> when this is the first value ever seen
 * by the group during a report run.
 *
 * @return <code>true</code> if this is the first value ever seen
 */
public boolean isFirstValue() { return firstValue; }

// ---------------- headers

/**
 * Adds a new header to the list of header sections.
 *
 * @param s a report section
 */
public void addHeader(Section s) {
    headerSections.add(s);
    s.setArea(Report.AREA_GROUP_HEADER);
}

/**
 * Adds a (possibly created) section after <var>afterThis</var> and
 * returns the section. If <var>section</var> is <code>null</code>,
 * a new section will be created.
 *
 * @param section the section to insert; if <code>null</code>, a new
 * section will be created
 * @param afterThis the new section will be inserted after this one
 */
public Section insertHeaderAfter(Section section, Section afterThis) {
    int i = headerSections.indexOf(afterThis);
    if (section == null)
	section = new Section(afterThis.getReport(),
			      I18N.get("Report.group_header"), false);
    headerSections.add(i + 1, section);
    section.setArea(Report.AREA_GROUP_HEADER);
    return section;
}

/**
 * Removes a header section from this group.
 *
 * @param s a report section
 */
public void removeHeader(Section s) {
    headerSections.remove(s);
}

/**
 * Returns an iterator over all header sections in the group.
 *
 * @return an iterator over the header sections
 */
public Iterator headers() { return headerSections.iterator(); }

// ---------------- footers

/**
 * Adds a new footer to the list of footer sections.
 *
 * @param s a report section
 */
public void addFooter(Section s) {
    footerSections.add(s);
    s.setArea(Report.AREA_GROUP_FOOTER);
}

/**
 * Adds a (possibly created) section after <var>afterThis</var> and
 * returns the section. If <var>section</var> is <code>null</code>,
 * a new section will be created.
 *
 * @param section the section to insert; if <code>null</code>, a new
 * section will be created
 * @param afterThis the new section will be inserted after this one
 */
public Section insertFooterAfter(Section section, Section afterThis) {
    int i = footerSections.indexOf(afterThis);
    if (section == null)
	section = new Section(afterThis.getReport(),
			      I18N.get("Report.group_footer"), false);
    footerSections.add(i + 1, section);
    section.setArea(Report.AREA_GROUP_FOOTER);
    return section;
}

/**
 * Removes a footer section from this group.
 *
 * @param s a report section
 */
public void removeFooter(Section s) {
    footerSections.remove(s);
}
/**
 * Returns an iterator over all footer sections in the group.
 *
 * @return an iterator over the footer sections
 */
public Iterator footers() { return footerSections.iterator(); }

/**
 * Returns <code>true</code> if the specified section is inside this group,
 * either as a header or a footer.
 *
 * @param s a section
 * @return <code>true</code> if the section is within this group
 */
public boolean contains(Section s) {
    return headerSections.contains(s) || footerSections.contains(s);
}

/**
 * Called by a report when it starts running, this method prepares the
 * group for use.
 */
public void reset() {
    value = null;
    newValue = firstValue = true;
    recordCount = 1;
}

/**
 * Writes this group as an XML tag. Asks each section to write itself
 * as well.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<group column=\"" + Utils.escapeXML(column.getId().toString())
		+ "\">");
    out.indent();

    out.println("<headers>");
    out.indent();
    for (Iterator iter = headerSections.iterator(); iter.hasNext(); )
	((Section)iter.next()).writeXML(out);
    out.outdent();
    out.println("</headers>");

    out.println("<footers>");
    out.indent();
    for (Iterator iter = footerSections.iterator(); iter.hasNext(); )
	((Section)iter.next()).writeXML(out);
    out.outdent();
    out.println("</footers>");

    out.outdent();
    out.println("</group>");
}

}
