package jimm.datavision;
import java.io.*;
//  import java.util.*;
import uk.co.wilson.xml.MinML2;
import org.xml.sax.*;

/**
 * A parameter reader reads an XML file and sets a report's parameter
 * values. This class is used when the report is being run from the
 * command line and the user has given us the name of an XML file containing
 * parameter elements.
 * <p>
 * Unlike a <code>ReportReader</code>, a parameter reader's constructor
 * takes not only the report but also the input method (file name, stream,
 * or reader). That way the report object doesn't have to know how to hold
 * on to those multipule input types.
 * 
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 * @see Report
 * @see ReportReader
 */
public class ParameterReader extends MinML2 {

/**
 * If there is no report element dtd-version attribute, this is the
 * default value to use.
 */
protected static final double DEFAULT_DTD_VERSION = 0.1;

protected Report report;
protected Parameter parameter;
protected String textData;
protected String fileName;
protected Reader inReader;
protected InputSource inInputSource;
protected InputStreamReader inInputStreamReader;

/**
 * Constructor.
 *
 * @param report the report whose parameters we are setting
 * @param fileName the parameter XML file name
 */
public ParameterReader(Report report, String fileName) {
    this.report = report;
    this.fileName = fileName;
}

/**
 * Constructor.
 *
 * @param report the report whose parameters we are setting
 * @param in the param XML reader
 */
public ParameterReader(Report report, Reader in) {
    this.report = report;
    inReader = in;
}

/**
 * Constructor. To specify a URL, use <code>new
 * InputSource("http://...")</code>.
 *
 * @param report the report whose parameters we are setting
 * @param in the param XML input source
 */
public ParameterReader(Report report, InputSource in) {
    this.report = report;
    inInputSource = in;
}

/**
 * Constructor.
 *
 * @param report the report whose parameters we are setting
 * @param in the param XML input stream reader
 */
public ParameterReader(Report report, InputStreamReader in) {
    this.report = report;
    inInputStreamReader = in;
}

/**
 * Returns the file name or, if that is <code>null</code>, the class name of
 * whatever input source was handed to a constructor.
 *
 * @param return a file name or class name
 */
public String getInputName() {
    if (fileName != null)
	return fileName;
    if (inReader != null)
	return "java.io.Reader";
    if (inInputSource != null)
	return "org.xml.sax.InputSource";
    if (inInputStreamReader != null)
	return "java.io.InputStreamReader";
    return "?";
}

/**
 * Reads parameter values from whichever input method was specified
 * in the constructor.
 */
public void read() throws FileNotFoundException, IOException, SAXException {
    if (fileName != null)
	parse(new InputStreamReader(new FileInputStream(fileName),
				    Report.XML_JAVA_ENCODING));
    else if (inReader != null)
	parse(inReader);
    else if (inInputSource != null)
	parse(inInputSource);
    else if (inInputStreamReader != null)
	parse(inInputStreamReader);
}

public void startElement(final String namespaceURI, final String localName,
			 final String qName, final Attributes attributes)
    throws SAXException
{
    // Get ready to start collecting text
    if (textData == null || textData.length() > 0)
	textData = new String();

    if ("parameter".equals(localName)) {
	String id = attributes.getValue("id");
	parameter = report.findParameter(id);
	if (parameter == null)
	    ErrorHandler.error(I18N.get("ParameterReader.unknown_id") + ' '
			       + id + ' '
			       + I18N.get("ParameterReader.in_xml"));
    }
}

public void endElement(final String namespaceURI, final String localName,
		       final String qName)
    throws SAXException
{
    if ("value".equals(localName) && parameter != null)
	parameter.addValue(textData);
    else if ("parameter".equals(localName))
	parameter = null;
}

/**
 * Reads text data. Text data inside a single tag can be broken up into
 * multiple calls to this method.
 */
public void characters(char ch[], int start, int length) {
    textData += new String(ch, start, length);
}

}
