package jimm.datavision;

/**
 * A suppression proc is an object used to decide if data should be
 * displayed or not. It returns <code>true</code> if the data should
 * be displayed or <code>false</code> if the data should be supressed
 * (should not be displayed).
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SuppressionProc implements Writeable {

protected Formula formula;
protected boolean hiding;

public SuppressionProc(Report report) {
    hiding = false;
    formula = new Formula(null, report, "");
}

public boolean isHidden() { return hiding; }
public void setHidden(boolean val) { hiding = val; }

/**
 * Returns formula used when not hiding.
 *
 * @return formula used when not hiding
 */
public Formula getFormula() { return formula; }

public boolean refersTo(Field f) {
    return formula.refersTo(f);
}

public boolean refersTo(Formula f) {
    return f == formula || formula.refersTo(f);
}

public boolean refersTo(UserColumn uc) {
    return formula.refersTo(uc);
}

public boolean refersTo(Parameter p) {
    return formula.refersTo(p);
}


/**
 * Returns <code>true</code> if the data should be suppressed (not displayed).
 * Returns <code>false</code> if the data should not be supressed (it should
 * be displayed).
 *
 * @return <code>true</code> if the data should be suppressed (not displayed)
 */
public boolean suppress() {
    if (hiding)
	return true;

    String expr = formula.getExpression();
    if (expr == null || expr.length() == 0)
	return false;

    Object obj = formula.eval();
    if (obj == null)		// Bogus Ruby code format (bad column)
	return false;
    return ((Boolean)obj).booleanValue();
}

/**
 * Writes this suppression proc as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    String expression = formula.getExpression();
    boolean hasFormula = expression != null && expression.length() > 0;

    if (!hiding && !hasFormula)
	return;

    out.print("<suppression-proc");

    if (hiding)
	out.print(" hide=\"true\"");

    if (hasFormula) {
	out.println(">");
	out.indent();
	formula.writeXML(out);
	out.outdent();
	out.println("</suppression-proc>");
    }
    else
	out.println(" />");
}

}
