package jimm.datavision.layout.pdf;
import jimm.datavision.*;
import java.io.File;
import java.io.PrintWriter;
import java.util.*;

/**
 * A PDF layout engine.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class PDFLE extends LayoutEngine implements LineDrawer {

protected static final String PDF_VERSION_STRING = "%PDF-1.0";

protected long loc;
protected ArrayList pages;
protected double linethickness;
protected ObjectID catalogID;
protected ObjectID pagesID;
protected ObjectID outlinesID;
protected ObjectID kidsID;
protected Page page;
protected XRef xref;
protected FontTracker fontTracker;

public PDFLE(PrintWriter out) {
    super(out);
    linethickness = 0;
    fontTracker = new FontTracker();
}

FontTracker getFontTracker() { return fontTracker; }

void println(String str) {
    out.println(str);
    loc += str.length() + File.separator.length();
}

void hereComesAnIndirectObject(ObjectID objID) {
    objID.setLoc(loc);
    xref.add(objID);
}

/**
 * Outputs the beginning of the document.
 */
protected void doStart() {
    loc = 0;
    pages = new ArrayList();
    catalogID = new ObjectID();
    pagesID = new ObjectID();
    outlinesID = new ObjectID();
    xref = new XRef();

    // Header
    println(PDF_VERSION_STRING);

    // We start the body with an informative, interesting comment
    println("% Generated by DataVision version " + info.Version);
    println("% " + info.URL);
}

/**
 * Outputs the end of the document. Generates cross-reference table (empty)
 * and trailer.
 */
protected void doEnd() {
    // The catalog and pages
    hereComesAnIndirectObject(catalogID);
    println(catalogID + " obj");
    println("<< /Type /Catalog /Pages " + pagesID + " R /Outlines "
	    + outlinesID + " R >>");
    println("endobj");

    hereComesAnIndirectObject(pagesID);
    println(pagesID + " obj");
    println("<< /Type /Pages /Count " +  pages.size() + " /Kids [");
    for (Iterator iter = pages.iterator(); iter.hasNext(); ) {
	Page p = (Page)iter.next();
	println(p.getObjectID() + " R");
    }
    println("] >>");
    println("endobj");

    hereComesAnIndirectObject(outlinesID);
    println(outlinesID + " obj");
    println("<< /Type /Outlines /Count 0 >>");
    println("endobj");

    // Output font(s)
    for (Iterator iter = fontTracker.fontObjs(); iter.hasNext(); ) {
	FontObj f = (FontObj)iter.next();
	hereComesAnIndirectObject(f.getID());
	f.output(this);
    }

    // Output pages
    int i = 0;
    for (Iterator iter = pages.iterator(); iter.hasNext(); ) {
	Page p = (Page)iter.next();
	println("% ======== Page " + (++i) + " ========");
	p.output();
    }

    // Cross reference
    long here = loc;		// Remember where we are for trailer
    xref.output(out);

    // Trailer
    out.println("trailer << /Size " + xref.size() + " /Root " + catalogID
		+ " R >>");
    out.println("startxref");
    out.println("" + here);
    out.println("%%EOF");	// End of file

    // Clean up pages
    for (Iterator iter = pages.iterator(); iter.hasNext(); ) {
	Page p = (Page)iter.next();
	p.cleanup();
    }

    pages =  null;
    xref = null;
}

/**
 * Start a new page.
 */
protected void doStartPage() {
    page = new Page(this, pagesID, pageWidth(), pageHeight());
    pages.add(page);
}

/**
 * End a page.
 */
protected void doEndPage() {
    page = null;
}

/**
 * Outputs a field.
 *
 * @param sect the field to output
 */
protected void doOutputField(Field field) {
    String fieldAsString = field.toString();
    if (fieldAsString == null || fieldAsString.length() == 0)
	return;

    Format format = field.getFormat() != null
	? field.getFormat() : Format.DEFAULT;

    page.addText(boundsOfField(field), fieldAsString, format);

//     // Borders
//     makeBorders(field);
}

protected Rectangle boundsOfField(Field f) {
    Rectangle bounds = new Rectangle(f.getBounds()); // Make a copy
    // x and y need to be the bottom left of the text box.
    if (currentSectionType != SECT_PAGE_FOOTER)
	bounds.y = (pageHeight() - pageHeightUsed) - bounds.y;
    else
	bounds.y = currentSection.getHeight() - bounds.y;
    bounds.y -= bounds.height;
    return bounds;
}


/**
 * Ignores image output
 *
 * @param image an image field
 */
protected void doOutputImage(ImageField field) {}

/**
 * Outputs a line. Calls <code>drawLine</code>.
 *
 * @param line a line
 * @see #drawLine
 */
protected void doOutputLine(Line line) {
    drawLine(line);
}

/**
 * Returns array of two strings that set/reset font size. If format is
 * <code>null</code> or has no size, return <code>null</code>.
 *
 * @return an array of two strings
 */
protected String[] selectFontSize(Format format) {
    if (format == null || format.getSize() == 0)
	return null;
    String[] size = new String[2];
    size[0] = "{\\fontsize{" + format.getSize() + "pt}{"
	+ (format.getSize() * 1.2) + "pt}";
    size[1] = "}";
    return size;
}

// /**
//  * Outputs borders.
//  */
// protected void makeBorders(Field field) {
//     if (field.getBorder() != null)
// 	field.getBorder().eachLine(this); // Calls drawLine for each line
// }

/**
 * Draw a single line.
 *
 * @param line a line
 */
public void drawLine(Line line) {
//     setLineThickness(line.getThickness());
//     putLine(line, xslope, yslope);
}

/**
 * Returns an array containing two doubleing-point values representing
 * the x and y values needed for a PDF line. PDF insists that
 * these two values be integers, but we use doubles for accuracy and
 * let <code>putLine</code> truncate them.
 *
 * @param slope a double
 * @return an array containing x and y values
 */
protected double[] pickNearestSlope(double slope) {
    boolean found = false;
    double xySlopeAbs = 0, xSlope = 0, ySlope = 0;
    for (int x = -6; x <= 6; ++x) {
	for (int y = -6; y <= 6; ++y) {
	    if (!found || xySlopeAbs > (slope - Math.abs(x / y))) {
		xySlopeAbs = Math.abs(slope - (x / y));
		xSlope = x;
		ySlope = y;
		found = true;
	    }
	}
    }
    if (!found) return null;

    double[] answer = new double[2];
    answer[0] = xSlope;
    answer[1] = ySlope;
    return answer;
}

/**
 * Sets the line thickness for drawing.
 *
 * @param t thickness
 */
protected void setLineThickness(double t) {
    if (linethickness != t) {
	linethickness = t;
	out.print("\\linethickness{" + linethickness + "pt}");
    }
}

/**
 * Outputs the PDF code that places the field on the page.
 */
protected void putField(Field field) {
    double y;
    Rectangle bounds = field.getBounds();
    if (currentSectionType == SECT_PAGE_FOOTER) {
	// Page footers are anchored to the bottom of the page
	y = bounds.y;
    }
    else {
	y = pageHeight() - (pageHeightUsed + bounds.y);
    }

    // In a \put command, x and y determine the bottom left corner of
    // whatever you are placing.
    y -= bounds.height;

    out.print("\\put(" + bounds.x + "," + y + "){");
}	    

/**
 * Outputs the PDF code that places the line on the page. Both xslope
 * and yslope are cast to ints because that's what PDF requires.
 *
 * @param line the line to draw
 * @param xslope the X slope of the line; must be from -6 through 6
 * @param yslope the Y slope of the line; must be from -6 through 6
 */
protected void putLine(Line line, double xslope, double yslope) {
    double y;
    Point p0 = line.getPoint(0);
    if (currentSectionType == SECT_PAGE_FOOTER) {
	// Page footers are anchored to the bottom of the page
	y = p0.y;
    }
    else {
	y = pageHeight() - (pageHeightUsed + p0.y);
    }

    // In a \put command, x and y determine the bottom left corner of
    // whatever you are placing.
    out.println("\\put(" + p0.x + "," + y + "){\\line(" + (int)xslope + ","
		+ (int)yslope + "){" + line.length() + "}}");
}	    

}
