package jimm.datavision.layout.pdf;
import jimm.datavision.Format;
import jimm.datavision.Rectangle;
import jimm.datavision.Utils;
import java.awt.Color;
import java.awt.Panel;
import java.awt.Font;
import java.awt.FontMetrics;
import java.io.File;
import java.util.*;

public class Page {

protected ObjectID id;
protected PDFLE pdf;
protected ObjectID parentID;
protected int pageWidth;
protected int pageHeight;
protected long len;
protected StringBuffer contents;
protected boolean prevWasText;
protected Color prevColor;
protected float[] colorComponents;
// protected double prevFontSize;
protected Panel panel;


public Page(PDFLE pdf, ObjectID parentID, double pgWidth, double pgHeight) {
    id = new ObjectID();

    this.pdf = pdf;
    this.parentID = parentID;
    pageWidth = (int)pgWidth;
    pageHeight = (int)pgHeight;
    contents = new StringBuffer();
    prevWasText = false;
    colorComponents = new float[4];
    panel = new Panel();
}

public ObjectID getObjectID() { return id; }

void output() {
    ObjectID contentsID = new ObjectID();

    // Start output
    pdf.hereComesAnIndirectObject(id);
    pdf.println(id.toString() + " obj");

    pdf.println("<<");		// Dictionary describing contents
    pdf.println("/Type /Page");
    pdf.println("/Parent " + parentID + " R");

    pdf.println("/Resources <<");

    pdf.println("/Font <<");
    for (Iterator iter = pdf.getFontTracker().fontObjsUsedBy(this);
	 iter.hasNext(); )
    {
	FontObj f = (FontObj)iter.next();
	f.outputResource(pdf);
    }
    pdf.println(">>");

    pdf.println("/ProcSet [/PDF /Text]");
    pdf.println(">>");		// End of resources

    pdf.println("/MediaBox [0 0 " + pageWidth + " " + pageHeight + "]");
    pdf.println("/Contents " + contentsID + " R");
    pdf.println(">>");		// End of dictionary describing contents
    pdf.println("endobj");

    // Output contents
    ObjectID lengthID = new ObjectID();
    pdf.hereComesAnIndirectObject(contentsID);
    pdf.println(contentsID + " obj");
    pdf.println("<< /Length " + lengthID + " R >>");
    pdf.println("stream");
    long contentsLength = outputContents();
    pdf.println("endstream");
    pdf.println("endobj");

    // Output contents length
    pdf.hereComesAnIndirectObject(lengthID);
    pdf.println(lengthID + " obj " + contentsLength + " endobj");
}

public void addText(Rectangle bounds, String str, Format format) {
    if (!prevWasText)
	contents.append("BT\n");

    if (format == null)
	format = Format.DEFAULT;

    setPDFFont(format);
    setPDFColor(format);
    setPDFLocation(str, format, bounds);
    stringIntoContents(str);

    prevWasText = true;
}

protected void setPDFFont(Format format) {
    Font f = format.getFont();
    FontTracker tracker = pdf.getFontTracker();
    if (!tracker.lastFontUsedIs(this, f))
	tracker.fontUsed(this, f).appendToBuf(contents, format.getSize());
}

protected void setPDFColor(Format format) {
    Color c = format.getColor();
    if (!c.equals(prevColor)) {
	if (c.equals(Color.black)) // A common case
	    contents.append("0 0 0 rg 0 0 0 RG ");
	else {
	    c.getColorComponents(colorComponents);
	    String colorStr = "" + colorComponents[0] + ' '
		+ colorComponents[1] + ' ' + colorComponents[2];
	    contents.append(colorStr + " rg " + colorStr + " RG ");
	}
	prevColor = c;
    }
}

protected void setPDFLocation(String str, Format format, Rectangle bounds) {
    contents.append("1 0 0 1 ");
    contents.append(alignedX(str, format, bounds));
    contents.append(' ');
    contents.append(bounds.y);	// Thanks, runako
    contents.append(" Tm");
}

protected double alignedX(String str, Format format, Rectangle bounds) {
    if (format.getAlign() == Format.ALIGN_LEFT)
	return bounds.x;

    FontMetrics fm = panel.getFontMetrics(format.getFont());
    int stringWidth = fm.stringWidth(str);

    double widthDiff = bounds.width - stringWidth;
    if (format.getAlign() == Format.ALIGN_CENTER)
	return bounds.x + widthDiff / 2;
    else			// Format.ALIGN_RIGHT
	return bounds.x + widthDiff;
}

protected long outputContents() {
    if (prevWasText)
	contents.append("ET\n");

    String str = contents.toString();
    pdf.println(str);
    return str.length() + File.separator.length();
}

public void cleanup() {
    contents = null;
}

/**
 * Shoves a string into the contents buffer.
 * <p>
 * We need to split the string into separate strings, one for each
 * linefeed or carriage return (or both).
 *
 * @param str the string we want to output
 */
protected void stringIntoContents(String str) {
    if (str == null || str.length() == 0)
	return;

    List strings = Utils.splitIntoLines(str);
    int numStrings = strings.size();

    contents.append(" (");
    if (numStrings == 1) {
	contents.append(makeSafe(str));
	contents.append(") Tj\n");
    }
    else {
	contents.append(makeSafe((String)strings.get(0)));
	contents.append(") Tj\n");
	for (int i = 1; i < numStrings; ++i) {
	    contents.append("(");
	    contents.append(makeSafe((String)strings.get(i)));
	    contents.append(") '\n");
	}
    }
}

/**
 * Returns a new string with all PDF special characters replaced by their
 * PDF-safe printable equivalents.
 *
 * @param str
 * @return a string safe for PDF output
 */
protected String makeSafe(String str) {
    StringBuffer buf = new StringBuffer();
    int len = str.length();
    for (int i = 0; i < len; ++i) {
	char c = str.charAt(i);
	// We handle linefeed and carriage return elsewhere. We don't
	// need to check for them here.
	switch (c) {
	case '\\': case '(': case ')':
	    buf.append('\\');
	    buf.append(c);
	    break;
	case '\t':
	    buf.append("\\t");
	    break;
	case '\b':
	    buf.append("\\b");
	    break;
	case '\f':
	    buf.append("\\f");
	    break;
	default:
	    buf.append(c);
	    break;
	}
    }
    return buf.toString();
}

}
