package jimm.datavision.source;
import jimm.datavision.*;
import java.util.*;

/**
 * A data source query.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Query implements Writeable {

public static final int SORT_UNDEFINED = -1;
public static final int SORT_DESCENDING = 0;
public static final int SORT_ASCENDING = 1;

protected Report report;
protected ArrayList joins;
protected String whereClause;
protected ArrayList sortCols;
protected ArrayList sortOrders;
protected ArrayList columns;	// Can't be a Set; we need column indices
protected ArrayList usercols;

/**
 * Constructor.
 *
 * @param report the report for which this query will retrieve data
 */
public Query(Report r) {
    report = r;
    joins = new ArrayList();
    whereClause = null;
    sortCols = new ArrayList();
    sortOrders = new ArrayList();
    columns = new ArrayList();
    usercols = new ArrayList();
}

/**
 * Returns <code>true</code> if the specified parameter exists within this
 * query's where clause.
 *
 * @param p a parameter
 * @return <code>true</code> if the specified parameter exists within
 * the where clause
 */
public boolean containsReferenceTo(Parameter p) {
    return whereClause != null && whereClause.indexOf(p.formulaString()) != -1;
}

/**
 * Adds a join to the list of joins used by this query.
 *
 * @param join a join
 */
public void addJoin(Join join) {
    joins.add(join);
}

/**
 * Adds all joins in the collection to our list.
 *
 * @param coll a collection of joins
 */
public void addAllJoins(Collection coll) {
    joins.addAll(coll);
}

/**
 * Removes a join from the list of joins used by this query.
 *
 * @param join a join
 */
public void removeJoin(Join join) {
    joins.remove(join);
}

/**
 * Removes all joins in the collection from our list.
 */
public void clearJoins() {
    joins.clear();
}

/**
 * Returns an iterator over all the joins used by this query.
 *
 * @return an iterator over all the joins
 */
public Iterator joins() { return joins.iterator(); }

/**
 * Returns the where clause fit for human consumption. This mainly means
 * that we substitute formula, parameter, and user column id numbers with
 * names. Called from a where clause editor. This code assumes that curly
 * braces are never nested.
 *
 * @return the eval string with formula, parameter, and user column id
 * numbers replaced with names
 * @see jimm.datavision.gui.sql.WhereClauseWin
 */
public String getEditableWhereClause() {
    return Expression.expressionToDisplay(report, whereClause);
}

/**
 * Returns the raw where clause string; may be <code>null</code>.
 *
 * @return the where clause string; may be <code>null</code>
 */
public String getWhereClause() {
    return whereClause;
}

/**
 * Sets the where clause (may be <code>null</code>). The string passed
 * to us contains parameter and formula display strings, not their
 * "real" <code>formulaString</code> representations. Translate the latter
 * into the former before saving this string.
 *
 * @param newWhereClause a where clause string; may be <code>null</code>
 */
public void setEditableWhereClause(String newWhereClause) {
    setWhereClause(Expression.displayToExpression(report, newWhereClause));
}

/**
 * Sets the where clause (may be <code>null</code>).
 *
 * @param newWhereClause a where clause string; may be <code>null</code>
 */
public void setWhereClause(String newWhereClause) {
    whereClause = newWhereClause;
}

/**
 * Adds the default sort order (descending) for the specified column.
 *
 * @param col a database column
 */
public void addSort(Column col) {
    addSort(col, SORT_DESCENDING);
}

/**
 * Adds a sort order for the specified column. The first character
 * of the <i>order</i> string is inspected. If it is a 'd' or 'D',
 * then it's taken to mean "descending". Anything else will result
 * in an "ascending" sort order.
 *
 * @param col a database column
 * @param order either <code>SORT_DESCENDING</code> or
 * <code>SORT_ASCENDING</code>
 */
public void addSort(Column col, int order) {
    sortCols.add(col);
    sortOrders.add(new Integer(order));
}

/**
 * Removes a sorting from the list.
 *
 * @param clause a where clause
 */
public void removeSort(Column col) {
    // I used to use a map, but then sorts would not keep their order,
    // which is important. That's why this code isn't just a map lookup
    // any more.
    for (int i = 0; i < sortCols.size(); ++i) {
	if (sortCols.get(i) == col) {
	    sortCols.remove(i);
	    sortOrders.remove(i);
	    return;
	}
    }
}

/**
 * Removes all sorts from our list.
 */
public void clearSorts() {
    sortCols = new ArrayList();
    sortOrders = new ArrayList();
}

/**
 * Returns an iterator over all columns.
 *
 * @return an iterator over all columns
 */
public Iterator columns() { return columns.iterator(); }

/**
 * Returns an iterator over all the sorted columns used by this query.
 *
 * @return an iterator
 */
public Iterator sortedColumns() { return sortCols.iterator(); }

/**
 * Returns the sort order (<code>SORT_DESCENDING</code>,
 * <code>SORT_ASCENDING</code>, or <code>SORT_UNDEFINED</code>) of the
 * specified column.
 *
 * @param col a database column
 * @return the sort order (<code>SORT_DESCENDING</code>,
 * <code>SORT_ASCENDING</code>, or <code>SORT_UNDEFINED</code>) of the
 * specified column.
 */
public int sortOrderOf(Column col) {
    // I used to use a map, but then sorts would not keep their order,
    // which is important. That's why this code isn't just a map lookup
    // any more.
    for (int i = 0; i < sortCols.size(); ++i) {
	if (sortCols.get(i) == col)
	    return ((Integer)sortOrders.get(i)).intValue();
    }
    return SORT_UNDEFINED;
}

/**
 * Returns the index of the specified column.
 *
 * @param column a database column
 */
public int indexOfColumn(Column column) {
    return columns.indexOf(column);
}

/**
 * Returns the index of the specified user column.
 *
 * @param uc a user column
 */
public int indexOfColumn(UserColumn uc) {
    return columns.size() + usercols.indexOf(uc);
}

/**
 * Builds collections of the columns actually used in the report.
 */
public void findColumnsUsed() {
    columns.clear();
    usercols.clear();
    report.withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof ColumnField) {
		Column col = ((ColumnField)f).getColumn();
		if (!columns.contains(col)) columns.add(col);
	    }
	    else if (f instanceof FormulaField) {
		FormulaField ff = (FormulaField)f;
		for (Iterator iter = ff.columnsUsed().iterator();
		     iter.hasNext(); )
		{
		    Column col = (Column)iter.next();
		    if (!columns.contains(col)) columns.add(col);
		}
		for (Iterator iter = ff.userColumnsUsed().iterator();
		     iter.hasNext(); )
		{
		    UserColumn uc = (UserColumn)iter.next();
		    if (!usercols.contains(uc)) usercols.add(uc);
		}
	    }
	    else if (f instanceof UserColumnField) {
		UserColumn uc = ((UserColumnField)f).getUserColumn();
		if (!usercols.contains(uc)) usercols.add(uc);
	    }
	}
	});

    // Add all columns used in sorts.
    for (Iterator iter = sortedColumns(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	if (!columns.contains(col)) columns.add(col);
    }
}

/**
 * Returns the number of columns in the query. Does not recalculate the
 * columns used; we assume this is being called after the qurey has been
 * run or <code>findColumnsUsed</code> has been called.
 */
public int getNumColumns() { return columns.size() + usercols.size(); }

/**
 * Called from <code>DataSource.reloadColumns/code>, this method gives the
 * query source a chance to tell its ancillary objects (such as joins and
 * the sort) to reload column objects.
 * <p>
 * This is necessary, for example, after a SQL database data source has
 * reloaded all of its table and column information. The old column
 * objects no longer exist. New ones (with the same ids, we assume) have
 * taken their place.
 *
 * @param dataSource the data source
 */
public void reloadColumns(DataSource dataSource)
{
    // Joins
    for (Iterator iter = joins(); iter.hasNext(); ) {
	Join j = (Join)iter.next();
	j.setFrom(dataSource.findColumn(j.getFrom().getId()));
	j.setTo(dataSource.findColumn(j.getTo().getId()));
    }

    // Sort columns
    ArrayList newSortCols = new ArrayList();
    for (Iterator iter = sortCols.iterator(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	newSortCols.add(dataSource.findColumn(col.getId()));
    }
    sortCols = newSortCols;

    // Columns. Some data sources recreate this list at the beginning of
    // every query execution, but I don't want to count on that.
    ArrayList newColumns = new ArrayList();
    for (Iterator iter = columns.iterator(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	newColumns.add(dataSource.findColumn(col.getId()));
    }
    columns = newColumns;
}

/**
 * Writes this query as an XML tag. Writes all joins, where clauses,
 * and sorts as well.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<query>");
    out.indent();

    for (Iterator iter = joins(); iter.hasNext(); )
	((Join)iter.next()).writeXML(out);

    if (whereClause != null && whereClause.length() > 0)
	out.println("<where><![CDATA[" + whereClause + "]]></where>");

    for (int i = 0; i < sortCols.size(); ++i) {
	int sortOrder = ((Integer)sortOrders.get(i)).intValue();
	out.println("<sort order=\""
		    + (sortOrder == Query.SORT_DESCENDING ? "desc" : "asc")
		    + "\" column=\""
		    + Utils.escapeXML(((Column)sortCols.get(i)).getId().toString())
		    + "\" />");
    }

    out.outdent();
    out.println("</query>");
}

}
