package jimm.datavision.test;
import jimm.datavision.Report;
import jimm.datavision.layout.CharSepLE;
import jimm.datavision.source.sql.Database;
import jimm.datavision.source.sql.SQLQuery;
import java.util.Iterator;
import java.io.File;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.sql.*;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.framework.Test;

/**
 * Tests the ability to give a connection to a report and the state of a
 * connection's query after reconnecting.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class ConnectionTest extends TestCase {

protected static final String EXAMPLE_REPORT =
    "jimm/datavision/test/test.xml";
protected static final String PARAMETER_XML_FILE_NAME =
    "jimm/datavision/test/test_parameters.xml";
protected static final String OUT_FILE =
    "/tmp/datavision_connection_test_out.txt";

protected static final String DRIVER_CLASS_NAME = "org.postgresql.Driver";
protected static final String CONNECTION_INFO =
    "jdbc:postgresql://127.0.0.1/dv_example";
protected static final String DB_NAME = "dv_example";
protected static final String DB_USER = "jimm";
protected static final String DB_PASSWORD = "";

public static Test suite() {
    return new TestSuite(ConnectionTest.class);
}

public ConnectionTest(String name) {
    super(name);
}

public void testConnection() {
    Connection conn = null;

    try {
	Driver d = (Driver)Class.forName(DRIVER_CLASS_NAME).newInstance();
	DriverManager.registerDriver(d);
	conn = DriverManager.getConnection(CONNECTION_INFO, DB_USER,
					   DB_PASSWORD);

	Report report = new Report();
	report.setDatabaseConnection(conn);

	File f = new File(OUT_FILE);
	f.deleteOnExit();
	PrintWriter out = new PrintWriter(new FileWriter(f));
	report.setLayoutEngine(new CharSepLE(out, '\t'));

	report.runReport();
    }
    catch (Exception e) {
	e.printStackTrace();
	fail("exception thrown: " + e);
    }
    finally {
	if (conn != null) {
	    try {
		conn.close();
	    }
	    catch (SQLException sqle) {
		fail("SQL exception thrown: " + sqle);
	    }
	}
    }
}

public void testQueryAfterReset() {
    Report report = new Report();
    try {
	report.setDatabasePassword(DB_PASSWORD);
	report.readFile(EXAMPLE_REPORT);

	Database db = (Database)report.getDataSource();
	SQLQuery query = (SQLQuery)db.getQuery();

	assertEquals("{jobs.ID} < 100", query.getWhereClause());
	assertEquals(4, tableCount(db));

	// We should only have two tables in the query.
	query.findColumnsUsed();
	assertEquals(2, query.getNumTables());

	db.reset(DRIVER_CLASS_NAME, CONNECTION_INFO, DB_NAME, DB_USER,
		 DB_PASSWORD);
	// The query doesn't have to be the same object, but it's where
	// clause (and all other information) should darned well be the same.
	assertEquals("{jobs.ID} < 100", query.getWhereClause());
	assertEquals(4, tableCount(db));

	// Make sure we still have two tables in the query.
	query.findColumnsUsed();
	assertEquals(2, query.getNumTables());
    }
    catch (Exception e) {
	fail(e.toString());
    }
}

protected int tableCount(Database db) {
    int i = 0;
    for (Iterator iter = db.tables(); iter.hasNext(); iter.next())
	++i;
    return i;
}

public static void main(String[] args) {
    junit.textui.TestRunner.run(suite());
}

}
