package jimm.datavision.test;
import jimm.datavision.Utils;
import jimm.datavision.Replacer;
import java.util.ArrayList;
import java.util.List;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.framework.Test;

public class UtilsTest extends TestCase {

public static Test suite() {
    return new TestSuite(UtilsTest.class);
}

public UtilsTest(String name) {
    super(name);
}

public void testEscapeXML() {
    assertEquals("", Utils.escapeXML(null));
    assertEquals("", Utils.escapeXML(""));
    assertEquals("abc&amp;def", Utils.escapeXML("abc&def"));
    assertEquals("abc&quot;def&apos;", Utils.escapeXML("abc\"def'"));
    assertEquals("&lt;&gt;", Utils.escapeXML("<>"));
}

public void testUnescapeXML() {
    assertEquals("", Utils.unescapeXML(null));
    assertEquals("", Utils.unescapeXML(""));
    assertEquals("abc&def", Utils.unescapeXML("abc&amp;def"));
    assertEquals("abc\"def'", Utils.unescapeXML("abc&quot;def&apos;"));
    assertEquals("<>", Utils.unescapeXML("&lt;&gt;"));
    assertEquals("abc def", Utils.unescapeXML("abc&#20;def"));
}

public void testNewlines() {
    assertEquals("", Utils.newlinesToXHTMLBreaks(null));
    assertEquals("", Utils.newlinesToXHTMLBreaks(""));
    assertEquals("\n<br />", Utils.newlinesToXHTMLBreaks("\n"));
    assertEquals("\n<br />", Utils.newlinesToXHTMLBreaks("\r"));
    assertEquals("\n<br />", Utils.newlinesToXHTMLBreaks("\r\n"));
    assertEquals("hello\n<br />world",
		 Utils.newlinesToXHTMLBreaks("hello\nworld"));
    assertEquals("hello\n<br />world",
		 Utils.newlinesToXHTMLBreaks("hello\rworld"));
    assertEquals("hello\n<br />world",
		 Utils.newlinesToXHTMLBreaks("hello\r\nworld"));
}

public void testSplit() {
    ArrayList typicalAnswer = new ArrayList();
    typicalAnswer.add("a");
    typicalAnswer.add("b");
    typicalAnswer.add("c");

    ArrayList origStringAnswer = new ArrayList();
    origStringAnswer.add("abc");

    // Special cases
    assertNull(Utils.split(null, null));
    assertEquals(origStringAnswer, Utils.split("abc", null));
    assertEquals(origStringAnswer, Utils.split("abc", "x"));

    assertEquals(typicalAnswer, Utils.split("a.b.c", "."));
    assertEquals(typicalAnswer, Utils.split("a**&*b**&*c", "**&*"));

    ArrayList answer = new ArrayList(typicalAnswer);
    answer.add(0, "");
    assertEquals(answer, Utils.split(".a.b.c", "."));
    assertEquals(answer, Utils.split("**&*a**&*b**&*c", "**&*"));

    answer = new ArrayList(typicalAnswer);
    answer.add(1, "");
    assertEquals(answer, Utils.split("a..b.c", "."));
    assertEquals(answer, Utils.split("a**&***&*b**&*c", "**&*"));

    answer = new ArrayList(typicalAnswer);
    answer.add("");
    assertEquals(answer, Utils.split("a.b.c.", "."));
    assertEquals(answer, Utils.split("a**&*b**&*c**&*", "**&*"));

    answer = new ArrayList(typicalAnswer);
    answer.add(1, "");
    answer.add("");
    assertEquals(answer, Utils.split("a..b.c.", "."));
    assertEquals(answer, Utils.split("a**&***&*b**&*c**&*", "**&*"));
}

public void testJoin() {
    assertEquals("", Utils.join(null, null));
    assertEquals("", Utils.join(null, ","));

    ArrayList list = new ArrayList();
    assertEquals("", Utils.join(list, null));
    assertEquals("", Utils.join(list, ","));

    list.add("a");
    assertEquals("a", Utils.join(list, null));
    assertEquals("a", Utils.join(list, ","));

    list.add("xyzzy");
    assertEquals("axyzzy", Utils.join(list, null));
    assertEquals("a,xyzzy", Utils.join(list, ","));
    assertEquals("a-*-xyzzy", Utils.join(list, "-*-"));
}

public void testReplace() {
    Replacer r = new Replacer() {
	public Object replace(String s) { return "x"; }
	};

    assertNull(Utils.replaceDelimited("", "", null, null));
    assertNull(Utils.replaceDelimited(null, null, null, null));
    assertEquals("abcde", Utils.replaceDelimited(null, null, r, "abcde"));
    assertEquals("abcde", Utils.replaceDelimited("", "", r, "abcde"));
    assertEquals("", Utils.replaceDelimited("{%", "}", r, ""));

    assertEquals("abxde", Utils.replaceDelimited("{%", "}", r, "ab{%c}de"));
    assertEquals("xabxdex", Utils.replaceDelimited("{%", "}", r,
						   "{%}ab{%c}de{%   }"));
    assertEquals("abcd{%e", Utils.replaceDelimited("{%", "}", r, "abcd{%e"));
    assertEquals("x.nil? ? nil : x / 100.0",
		 Utils.replaceDelimited("{", "}", r,
					"{jobs.hourly rate}.nil? ? nil : {jobs.hourly rate} / 100.0"));
}

public void testReplaceNotAfter() {
    Replacer r = new Replacer() {
	public Object replace(String s) { return "x"; }
	};

    assertNull(Utils.replaceDelimited("#", "", "", null, null));
    assertNull(Utils.replaceDelimited("#", "#", "#", null, null));
    assertNull(Utils.replaceDelimited("#", null, null, null, null));
    assertEquals("abcde", Utils.replaceDelimited("#", null, null, r, "abcde"));
    assertEquals("abcde", Utils.replaceDelimited("#", "", "", r, "abcde"));
    assertEquals("", Utils.replaceDelimited("#", "{%", "}", r, ""));

    assertEquals("abxde",
		 Utils.replaceDelimited("#", "{%", "}", r, "ab{%c}de"));
    assertEquals("xabxdex", Utils.replaceDelimited("#", "{%", "}", r,
						   "{%}ab{%c}de{%   }"));
    assertEquals("abcd{%e",
		 Utils.replaceDelimited("#", "{%", "}", r, "abcd{%e"));

    assertEquals("ab#{c}de",
		 Utils.replaceDelimited("#", "{", "}", r, "ab#{c}de"));
    assertEquals("ab##{c}de",
		 Utils.replaceDelimited("#", "{", "}", r, "ab##{c}de"));
    assertEquals("ab##{x}dex",
		 Utils.replaceDelimited("#", "{", "}", r, "ab##{{c}}de{z}"));
    assertEquals("ab#{x}de",
		 Utils.replaceDelimited("#", "{", "}", r, "ab#{{c}}de"));
    assertEquals("ab#{x}de",
		 Utils.replaceDelimited("#", "{%", "}", r, "ab#{{%c}}de"));
}

public void testLineSplit() {
    ArrayList answer = new ArrayList();
    answer.add("");

    List split = Utils.splitIntoLines("");
    assertEquals(answer.size(), split.size());
    assertEquals(answer, split);

    // The three flavors of line endings
    String[] endings = { "\n", "\r", "\r\n" };

    // We want to see an array with one empty string (the text before the
    // newline). There is no string after the newline.
    for (int i = 0; i < endings.length; ++i) {
	split = Utils.splitIntoLines(endings[i]);
	assertEquals(answer.size(), split.size());
	assertEquals(answer, split);
    }

    answer.add("x");
    for (int i = 0; i < endings.length; ++i) {
	split = Utils.splitIntoLines(endings[i] + "x");
	assertEquals(answer.size(), split.size());
	assertEquals(answer, split);
    }

    answer = new ArrayList();
    answer.add("line one");
    answer.add("part deux");
    answer.add("three's a crowd, eh?");
    answer.add("");

    split = Utils.splitIntoLines("line one\npart deux\nthree's a crowd, eh?"
				 + "\n\n");
    assertEquals(answer.size(), split.size());
    assertEquals(answer, split);
}

public static void main(String[] args) {
    junit.textui.TestRunner.run(suite());
}

}
