package jimm.datavision.testdata;
import java.io.FileReader;
import uk.co.wilson.xml.MinML2;
import org.xml.sax.*;

/**
 * Skeleton for creating a schema.sql file by reading an XML description
 * of a schema.
 * 
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */

public abstract class SchemaGen extends MinML2 {

protected String tableName;
protected String columnName;
protected String type;
protected int size;
protected boolean notNull;
protected boolean primaryKey;
protected boolean isFirstColumn;

public void run(String schemaXMLFile) {
    try {
	parse(new FileReader(schemaXMLFile));
    }
    catch (Exception e) {
	e.printStackTrace();
	System.exit(1);
    }
}

public void startElement(final String namespaceURI, final String localName,
			 final String qName, final Attributes attributes)
    throws SAXException
{
    if ("table".equals(localName)) table(attributes);
    else if ("column".equals(localName)) column(attributes);
}

public void endElement(final String namespaceURI, final String localName,
		       final String qName)
    throws SAXException
{
    if ("table".equals(localName)) endTable();
}

/**
 * Parses a table XML tag and calls <code>makeTable</code>.
 *
 * @param line a line of XML
 * @see #makeTable
 */
protected void table(Attributes attributes) {
    tableName = attributes.getValue("name");
    makeTable();
    isFirstColumn = true;
}

/**
 * Parses a column XML tag and calls <code>printColumn</code>. Also handles
 * commas and indentation.
 *
 * @param line a line of XML
 * @see #makeTable
 */
protected void column(Attributes attributes) {
    columnName = attributes.getValue("name");
    type = attributes.getValue("type");
    String sizeStr = attributes.getValue("size");
    if (sizeStr == null)
	size = 0;
    else
	size = Integer.parseInt(sizeStr);
    notNull = (attributes.getValue("not-null") != null);
    primaryKey = (attributes.getValue("primary-key") != null);

    if (!isFirstColumn)
	System.out.print(",");
    System.out.println();
    System.out.print("\t");

    printColumn();
    isFirstColumn = false;
}

/**
 * Outputs the SQL needed to create a database table. Optionally prints
 * the SQL needed to destroy the table first.
 */
protected abstract void makeTable();

/**
 * Outputs the SQL needed to close a create table statement.
 */
protected abstract void endTable();

/**
 * Prints the SQL needed to create a database column within a create table
 * statement.
 */
protected void printColumn() {
    if (columnName.indexOf(' ') >= 0
	|| !columnName.equals(columnName.toLowerCase()))
	System.out.print("\"" + columnName + "\"");
    else
	System.out.print(columnName);
    printType();
    if (notNull) printNotNull();
    if (primaryKey) printPrimaryKey();
}

/**
 * Prints the SQL that defines a column's type.
 */
protected abstract void printType();

/**
 * Prints the SQL that defines a column as NOT NULL.
 */
protected abstract void printNotNull();

/**
 * Prints the SQL that defines a column as a primary key.
 */
protected abstract void printPrimaryKey();

}
